// Grid.cpp : implementation file
//

#include "stdafx.h"
#include "Boggle.h"
#include "Grid.h"

extern	CString *Library;
extern	int		LibLen;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CGrid

// Historical high water marks (only need one copy for all grids):
int	CGrid::bestEverWords = 0;
int	CGrid::bestEverLetters = 0;


CGrid::CGrid()
	: letterCnt(0), wordCnt(0),
	  randomLetter('A','Z')
{
	int i, j;
	// Initialize the fitnessGrid (only the center is used, border is ignored):
	for (i=1; i<=ROWS; i++)
		for (j=1; j<=COLS; j++)
			fitnessGrid[i][j] = 0;
	// Initialize the letterGrid:
	// Put in the borders (the center will be overwritten after sex):
	for (j=0; j<=COLS+1; j++)				// Top row
		letterGrid[0][j] = '\0';
	for (i=1; i<=ROWS; i++)					// Left side
		letterGrid[i][0] = '\0';
	for (i=1; i<=ROWS; i++)					// Right side
		letterGrid[i][COLS+1] = '\0';
	for (j=0; j<=COLS+1; j++)				// Bottom row
		letterGrid[ROWS+1][j] = '\0';
}

CGrid::~CGrid()
{
}

BEGIN_MESSAGE_MAP(CGrid, CStatic)
	//{{AFX_MSG_MAP(CGrid)
		// NOTE - the ClassWizard will add and remove mapping macros here.
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


// The copy ctor:
CGrid::CGrid(const CGrid &copy)
{
	for (int r=0; r<ROWS+2; r++)
		for (int c=0; c<COLS+2; c++)
		{
			fitnessGrid[r][c] = copy.fitnessGrid[r][c];
			letterGrid[r][c] = copy.letterGrid[r][c];
		}
	wordCnt = copy.wordCnt;
	letterCnt = copy.letterCnt;
}

// The (=) assignment:
const CGrid& CGrid::operator=(const CGrid &right)
{
	if (&right == this)
		return *this;

	for (int r=0; r<ROWS+2; r++)
		for (int c=0; c<COLS+2; c++)
		{
			fitnessGrid[r][c] = right.fitnessGrid[r][c];
			letterGrid[r][c] = right.letterGrid[r][c];
		}
	wordCnt = right.wordCnt;
	letterCnt = right.letterCnt;
	
	return *this;
}

// Be fruitful and multiply:
CGrid CGrid::operator *(const CGrid &g)
{
	CGrid	child;		// The ctor puts the sentinel border on.
						// Don't initialize center, it will be overwritten.

	// OK, let's have sex!
	// The _this_ grid is the Mommy (since the multiplication
	// occurs in her body) and the passed grid (g) is the Daddy.
	for (int i=1; i<=ROWS; i++)
	{
		for (int j=1; j<=COLS; j++)
		{
			// Clean out the child fitness grid because FindString()
			// uses additive procedure.  (letterGrid[][] doesn't
			// need cleaning.  It is overwritten.)
			child.fitnessGrid[i][j] = 0;

			const int momFitness = fitnessGrid[i][j];
			const int dadFitness = g.fitnessGrid[i][j];
			if ((dadFitness == 0) && (momFitness == 0))
			{
				// Bad genes. Mutate (choose another letter).  Occaisionally
				// the mutation will return one of the parent letters.  So?
				child.letterGrid[i][j] = randomLetter.Pick();
			}
			else
				child.letterGrid[i][j] = 
					((momFitness > dadFitness) ?
					letterGrid[i][j] : g.letterGrid[i][j]);
		}
	}
	// Deliver this baby:
	return child;
}

void CGrid::Init()
{
	// Fill the center with random characters:
	for (int i=1; i<=ROWS; i++)
		for (int j=1; j<=COLS; j++)
			letterGrid[i][j] = randomLetter.Pick();
}

int CGrid::DetermineFitness()
{
	for (int i=0; i<::LibLen; i++)			// Look for all strings
	{
		CString s = ::Library[i];
		if (FindString(s) == FOUND)
			wordCnt++;						// Found one
	}
	// Log the grid if it's better than the historical highs:
	if ((wordCnt > bestEverWords) ||
		((wordCnt == bestEverWords) && (letterCnt > bestEverLetters)))
	{
		bestEverWords = wordCnt;
		bestEverLetters = letterCnt;
//		Log();
	}
	return wordCnt;
}

int CGrid::GetLetterCount(void) const
{
	return letterCnt;
}

FIND CGrid::FindString(CString s)
{
	for (int i=1; i<=ROWS; i++)				// Look in rows 1-8
		for (int j=1; j<=COLS; j++)			// Look in cols 1-8
			if (FindString(s, i, j) == FOUND)
				return FOUND;
	return NOTFOUND;
}

FIND CGrid::FindString(CString s, int row, int col)
{
	if (letterGrid[row][col] != s[0])
		return NOTFOUND;

	// Found a letter!
	
	// But don't celebrate yet.  It's not a word unless it's the last letter:
	if (s.GetLength() == 1)
	{
		// Found a word!  Now we can celebrate.  Start tallying points:
		letterCnt++;					// How's the grid doing? (for reporting)
		fitnessGrid[row][col]++;		// How's the character doing?  (used during sex)
		// Unwind the recursion:
		return FOUND;
	}

	// The passed string MIGHT start here.  Look around for the rest.
	s = s.Mid(1);				// Get the rest of the string
	if (FindString(s, row-1, col) == FOUND)		// Look north
	{
		letterCnt++;
		fitnessGrid[row][col]++;		// The letters along the path are rewarded
		return FOUND;
	}
	if (FindString(s, row-1, col+1) == FOUND)		// Look northeast
	{
		letterCnt++;
		fitnessGrid[row][col]++;		// The letters along the path are rewarded
		return FOUND;
	}
	if (FindString(s, row, col+1) == FOUND)		// Look east
	{
		letterCnt++;
		fitnessGrid[row][col]++;		// The letters along the path are rewarded
		return FOUND;
	}
	if (FindString(s, row+1, col+1) == FOUND)		// Look southeast
	{
		letterCnt++;
		fitnessGrid[row][col]++;		// The letters along the path are rewarded
		return FOUND;
	}
	if (FindString(s, row+1, col) == FOUND)		// Look south
	{
		letterCnt++;
		fitnessGrid[row][col]++;		// The letters along the path are rewarded
		return FOUND;
	}
	if (FindString(s, row+1, col-1) == FOUND)		// Look southwest
	{
		letterCnt++;
		fitnessGrid[row][col]++;		// The letters along the path are rewarded
		return FOUND;
	}
	if (FindString(s, row, col-1) == FOUND)		// Look west
	{
		letterCnt++;
		fitnessGrid[row][col]++;		// The letters along the path are rewarded
		return FOUND;
	}
	if (FindString(s, row-1, col-1) == FOUND)		// Look northwest
	{
		letterCnt++;
		fitnessGrid[row][col]++;		// The letters along the path are rewarded
		return FOUND;
	}

	// The first part of the string was looking good, but it didn't pan out:
	return NOTFOUND;
}

void CGrid::Log()
{
	// Put this grid out to the display:

	// This loop was done once already (in DetermineFitness), but which words were
	// found was not stored, so let's redo it here and log the results:
//	letterCnt = 0;							// Reset so FindString doesn't overreport
//	int wCnt = 0;
//	for (i=0; i<LIB_COUNT; i++)				// Look for all strings
//	{
//		char *s = ::library[i];
//		if (FindString(s) == FOUND)
//		{
//			// Found one:
//			wCnt++;
//			outFile << wCnt << ": " << s << "\n";
//		}
//	}
//	outFile << "\n****************************************\n";
//	outFile.close();
}

